/*    Copyright (C) 2005 XIAO, Gang of Universite de Nice - Sophia Antipolis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

	/* library functions for sysmask. */

  /* dos/mac to unix/linux translation */
void _tolinux(char *p)
{
    char *pp,*p1;
    pp=strchr(p,13); if(pp==NULL) return;
    for(p1=pp; *pp; pp++) {
	if(*pp==13) {
	    if(*(pp+1)=='\n' || (pp>p && *(pp-1)=='\n') ||
	       (*(pp+1)=='\\' && *(pp+2)=='\n')) continue;
	    else *pp='\n';
	}
	*p1++=*pp;
    }
    *p1=0;
}

	/* optimized and secure strcpy */
void mystrncpy(char *dest, char *src, int lim)
{
    int i;
    i=strlen(src); if(i<0) i=0;
    if(i>=lim) i=lim-1;
    memmove(dest,src,i); dest[i]=0;
}

	/* Points to the end of the word */
char *find_word_end(char *p)
{
    while(!myisspace(*p) && *p!=0) p++;
    return p;
}

	/* Strips leading spaces */
char *find_word_start(char *p)
{
    while(myisspace(*p)) p++;
    return p;
}

	/* Points to the end of a line */
char *find_line_end(char *p)
{
    char *pp=strchr(p,'\n');
    if(pp==NULL) pp=p+strlen(p);
    return pp;
}

	/* strip trailing spaces; return string end. */
char *strip_trailing_spaces(char *p)
{
    char *pp;
    if(*p==0) return p;
    for(pp=p+strlen(p)-1; pp>=p && myisspace(*pp); pp--);
    if(pp[1]) pp[1]=0; return pp;
}

	/* For independence to libc. */
char *mystrstr(const char *p1, const char *p2)
{
    char *pp;
    int l = strlen(p2);
    for(pp=strchr(p1,*p2); pp; pp=strchr(++pp,*p2)) {
	if(strncmp(pp,p2,l)==0) return pp;
    }
    return NULL;
}

	/* Find first occurrence of word */
char *wordchr(char *p, char *w)
{
    char *r; int n;

    if(*w==0) return NULL;
    n=strlen(w);
    for(r=mystrstr(p,w);r!=NULL && 
	( (r>p && !myisspace(*(r-1))) || (!myisspace(*(r+n)) && *(r+n)!=0) );
	r=mystrstr(r+1,w));
    return r;
}

	/* make a string from components.
	 * Returns 0 if OK. -1 if overflow. */
int makestr(char buf[], const char *comp[], long size)
{
    int i;
    long l, k;
    for(i=0, l=0; i<256 && comp[i]; i++) {
	k=strlen(comp[i]); if(k+l>=size) {
	    buf[0]=0; return -1;
	}
	memmove(buf+l,comp[i],k); l+=k;
    }
    buf[l]=0; return 0;
}

int str2int(char *p)
{
    int i,j,k;
    p=find_word_start(p);
    if(*p=='-') j=-1; else j=1;
    for(k=i=0;myisdigit(*p) && k<8;k++,p++) {
	i*=10; i+=*p-'0';
    }
    return i*j;
}

	/* searches a list. Returns index if found, -1 if nomatch. 
	 * Uses binary search, list must be sorted. */
int search_list(void *list, int items, size_t item_size, const char *str)
{
    int i1,i2,j,k;
    char **p;
    char c;
    
    if(items<=0) return -1;
    j=0; c=*str;
    p=list;
    k=**p-c; if(k==0) k=strcmp(*p,str);
    if(k==0) return k; if(k>0) return -1;
    p=list+(items-1)*item_size; 
    k=**p-c; if(k==0) k=strcmp(*p,str);
    if(k==0) return items-1; if(k<0) return ~items;
    for(i1=0,i2=items-1;i2>i1+1;) {
	j=(i2+i1)/2;
	p=list+(j*item_size);
	k=**p-c; if(k==0) k=strcmp(*p,str);
	if(k==0) return j;
	if(k>0) {i2=j; continue;}
	if(k<0) {i1=j; continue;}	
    }
    return ~i2;
}

int gettrigger(char *s)
{
    int i;

    s=find_word_start(s); if(!*s) return 0;
    if(myisdigit(*s)) {
	i=str2int(s); if(i>0) return i;
    }
    else {
	i=search_list(triggers,trcnt,sizeof(triggers[0]),s);
	if(i>=0) return triggers[i].val;
    }
    return 0;
}

void addmask(struct sysmask_struct *dest, struct sysmask_struct *src)
{
    int i,j;
    if(src->trigger) {
	if(dest->trigger) {
	    if(dest->trigger!=src->trigger) {
		dest->active.m1|=dest->triggered.m1;
		dest->active.m2|=dest->triggered.m2;
		dest->active.cap|=dest->triggered.cap;
		memset(&dest->triggered,0,sizeof(struct sysmask_maskset));
	    }
	    else {
		if(dest->tcnt<src->tcnt)
		  src->tcnt=dest->tcnt;
	    }
	}
	dest->trigger=src->trigger;
	dest->tcnt=src->tcnt;
	dest->triggered.m1|=src->triggered.m1;
	dest->triggered.m2|=src->triggered.m2;
	dest->triggered.cap|=src->triggered.cap;
    }
    for(i=0;i<SMK_EXCEPTS;i++) {
	if(src->except[i]>NR_sysmasks && src->except[i]!=SMKD_EXCEPT) {
	    src->except[i]=0;
	}
	if(src->except[i]!=0) {
	    for(j=0;j<SMK_EXCEPTS && src->except[i]!=dest->except[j];j++);
	    if(j>=SMK_EXCEPTS) {
		if(src->except[i]<=NR_sysmasks) {
		    if((sysmask_tab[src->except[i]-1]&dest->active.m1)!=0)
		      src->except[i]=0;
		}
		else if(dest->active.m1) src->except[i]=0;
	    }
	}
    }
    memmove(&dest->except,&src->except,SMK_EXCEPTS*sizeof(short));
    dest->active.m1|=src->active.m1;
    dest->active.m2|=src->active.m2;
    dest->active.cap|=src->active.cap;
}

void join_excepts(unsigned short int dest[], unsigned short int src[])
{
    int i,j;
    for(i=0;i<SMK_EXCEPTS && src[i]; i++) {
	for(j=0; j<SMK_EXCEPTS; j++) {
	    if(dest[j]==0) {dest[j++]=src[i]; break;}
	    if(dest[j]==src[i]) break;
	}
	if(j>=SMK_EXCEPTS) break;
    }
}

void onemask(char *name, struct sysmask_maskset *mk,unsigned short int ex[])
{
    int i,k;
    if(*name=='x') {
	if(ex==NULL) return;
	i=gettrigger(name+1)+1;
	if(i<=0 || (i>NR_sysmasks && i!=SMKD_EXCEPT)) return;
	if(i==1 && name[1]!='0' && strcmp(name+1,syscall_zero)!=0) return;
	for(k=0; k<SMK_EXCEPTS && ex[k]!=0; k++) {
	    if(ex[k]==i) return;
	}
	if(k<SMK_EXCEPTS) ex[k]=i;
	return;
    }
    i=search_list(maskname,maskcnt,sizeof(maskname[0]),name);
    if(i<0) {
	if(strncmp(name,"no",2)==0) {
	    i=search_list(maskname,maskcnt,sizeof(maskname[0]),name+2);
	    if(i>=0) {
		mk->m1&=~maskname[i].m1; mk->m2&=~maskname[i].m2;
		mk->cap&=~maskname[i].cap;
	    }
	}
	return;
    }
    mk->m1|=maskname[i].m1; mk->m2|=maskname[i].m2;
    mk->cap|=maskname[i].cap;
}

	/* modify a string. Bufferlen must be ast least SMK_PATH_MAX */
	/* returns 0 if OK. */
int namestr_modify(char *start, char *bad_beg, char *bad_end, char *good)
{
    int ln, le;
    
    ln=strlen(good); le=strlen(bad_end);
    if(ln+le+(bad_beg-start)>=SMK_PATH_MAX) {
	*start=0; return -1;
    }
    if(ln!=bad_end-bad_beg) memmove(bad_beg+ln,bad_end,le+1);
    memmove(bad_beg,good,ln);
    return 0;
}

struct {
    char *name;
    char *stars;
} syscall_names[NR_sysmasks];

	/* get the syscall names according to trigger table */
void get_syscall_names(void)
{
    int i, j;
    memset(syscall_names,0,sizeof(syscall_names));
    for(i=0;i<trcnt;i++) {
	j=triggers[i].val;
	if(j<0 || j>=NR_sysmasks) continue;
	syscall_names[j].name=triggers[i].name;
	syscall_names[j].stars="";
	if(sysmask_tab[j] & SM1_MINIMAL) {
	    syscall_names[j].stars="******"; continue;
	}
	if(sysmask_tab[j] & SM1_NONSTD) {
	    syscall_names[j].stars="*****"; continue;
	}
	if((sysmask_tab[j] & SM1_NONFREQ) ||
	   j==__NR_modify_ldt
	   ) {
	    syscall_names[j].stars="****"; continue;
	}
	if((sysmask_tab[j] & (SM1_RESOURCE|SM1_PRIV|SM1_BRK)) ||
	   j==__NR_execve || j==__NR_ioctl || 
	   j==__NR_setuid || j==__NR_setgid || j==__NR_setreuid ||
	   j==__NR_setregid || j==__NR_setgroups ||
	   j==__NR_setuid32 || j==__NR_setgid32 || j==__NR_setreuid32 ||
	   j==__NR_setregid32 || j==__NR_setgroups32
	   ) {
	    syscall_names[j].stars="***"; continue;
	}
	if((sysmask_tab[j] & 
	   (SM1_DREAD|SM1_DWRITE|SM1_CHMOD|SM1_OPEN)) &&
	   j!=__NR_fchmod && j!=__NR_fstat && j!=__NR_getdents &&
	   j!=__NR_getcwd && j!=__NR_fstat64 && j!=__NR_getdents64 &&
	   j!=__NR_fsetxattr && j!=__NR_fgetxattr && j!=__NR_flistxattr &&
	   j!=__NR_fremovexattr
	   ) {
	    syscall_names[j].stars="**"; continue;
	}
	if((sysmask_tab[j] & ~SM1_HARMLESS) ||
	   j==__NR_pread64 ||
	   j==__NR_pwrite64 || j==__NR_read || j==__NR_write || 
	   j==__NR_readv || j==__NR_writev
	   ) {
	    syscall_names[j].stars="*"; continue;
	}
    }
    syscall_names[0].name=syscall_zero;
    syscall_names[0].stars="******";
    syscall_names[__NR_sysmask].name="sysmask";
}

